<?php

namespace Drupal\entity_hierarchy\Plugin\views\argument;

use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\entity_hierarchy\Storage\QueryBuilderFactory;
use Drupal\views\Plugin\views\argument\ArgumentPluginBase;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Defines a base argument plugin.
 */
abstract class EntityHierarchyArgumentPluginBase extends ArgumentPluginBase {

  /**
   * Constructs a new HierarchyIsParentOfEntity object.
   *
   * @param array $configuration
   *   Configuration.
   * @param string $plugin_id
   *   Plugin ID.
   * @param mixed $plugin_definition
   *   Definition.
   * @param \Drupal\entity_hierarchy\Storage\QueryBuilderFactory $queryBuilderFactory
   *   Query builder factory.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   Entity type manager.
   */
  final public function __construct(
    array $configuration,
    $plugin_id,
    $plugin_definition,
    protected QueryBuilderFactory $queryBuilderFactory,
    protected EntityTypeManagerInterface $entityTypeManager,
  ) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('entity_hierarchy.query_builder_factory'),
      $container->get('entity_type.manager'),
    );
  }

  /**
   * {@inheritdoc}
   */
  public function buildOptionsForm(&$form, FormStateInterface $form_state): void {
    $options = range(1, 50);
    $form['depth'] = [
      '#type' => 'select',
      '#empty_option' => $this->t('No restriction'),
      '#empty_value' => 0,
      '#options' => array_combine($options, $options),
      '#title' => $this->t('Depth'),
      '#default_value' => $this->options['depth'],
      '#description' => $this->t('Filter to children that are at most this many levels deeper than their parent. E.g. for immediate children, select 1.'),
    ];
    parent::buildOptionsForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  protected function defineOptions() : array {
    $options = parent::defineOptions();

    // Allow filtering depth as well.
    $options['depth'] = ['default' => 0];

    return $options;
  }

  /**
   * Loads the parent entity from the argument.
   *
   * @return \Drupal\Core\Entity\ContentEntityInterface|null
   *   Parent entity if exists.
   */
  protected function loadEntity(): ?ContentEntityInterface {
    $entity = $this->entityTypeManager->getStorage($this->getEntityType())->load($this->argument);
    assert(is_null($entity) || $entity instanceof ContentEntityInterface);
    return $entity;
  }

}
